
#define ARM7

/*---------------------------------------------------------------------------------
	$Id: interruptDispatcher.s,v 1.7 2006/04/26 05:11:31 wntrmute Exp $

	Copyright (C) 2005
		Dave Murphy (WinterMute)

	This software is provided 'as-is', without any express or implied
	warranty.  In no event will the authors be held liable for any
	damages arising from the use of this software.

	Permission is granted to anyone to use this software for any
	purpose, including commercial applications, and to alter it and
	redistribute it freely, subject to the following restrictions:

	1.	The origin of this software must not be misrepresented; you
		must not claim that you wrote the original software. If you use
		this software in a product, an acknowledgment in the product
		documentation would be appreciated but is not required.
	2.	Altered source versions must be plainly marked as such, and
		must not be misrepresented as being the original software.
	3.	This notice may not be removed or altered from any source
		distribution.

	$Log: interruptDispatcher.s,v $
	Revision 1.7  2006/04/26 05:11:31  wntrmute
	rebase dtcm, take __irq_flags and __irq_vector from linker script
	move arm7 irq vector & irq flags to actual locations
	
	Revision 1.6  2006/04/23 18:19:15  wntrmute
	reworked interrupt code to allow dtcm moving
	
	Revision 1.5  2005/12/12 13:01:55  wntrmute
	disable interrupts on return from user handler
	
	Revision 1.4  2005/10/21 22:43:42  wntrmute
	restore REG_IME on exit from null handler
	
	Revision 1.3  2005/09/27 18:21:53  wntrmute
	safer nested interrupt support
	
	Revision 1.2  2005/09/04 16:37:01  wntrmute
	check for NULL handler
	
	Revision 1.1  2005/09/03 17:09:35  wntrmute
	added interworking aware interrupt dispatcher
	

---------------------------------------------------------------------------------*/

#ifdef ARM7
	.text
#endif

#ifdef ARM9
	.section	.itcm,"ax",%progbits
#endif

	.extern	_irqTableSimple
	.code 32

	.global	_IntrMain
@---------------------------------------------------------------------------------
_IntrMain:
@---------------------------------------------------------------------------------
	mrs	r0, spsr
	stmfd	sp!, {r0,lr}	@ {spsr, lr}

	mov	r3, #0x4000000		@ REG_BASE
	
	ldr	r1, [r3,#0x210]		@ REG_IE
	ldr	r2, [r3,#0x214]		@ REG_IF
	and	r1,r1,r2

#ifdef ARM7
	ldr	r2, [r3, #-8]		@\mix up with BIOS irq flags at 3007FF8h,
	orr	r2, r2, r1		@ aka mirrored at 3FFFFF8h, this is required
	str	r2, [r3, #-8]		@/when using the (VBlank)IntrWait functions
#endif

#ifdef ARM9
@	ldr	r0,=__irq_flags		@ defined by linker script (new ver)
	ldr	r0,=0x00803FF8		@ Bios irq flags on ARM9 (old ver)

	ldr	r2,[r0]
	orr	r2,r2,r1
	str	r2,[r0]
#endif

	ldr	r2,=_irqTableSimple
@---------------------------------------------------------------------------------
findIRQ:
@---------------------------------------------------------------------------------
	ands r0, r1, #0x20 @ IRQ_TIMER2 = BIT(5)
	bne	got_handler_mi
	add	r2, r2, #4
	ands r0, r1, #0x10 @ IRQ_TIMER1 = BIT(4)
	bne	got_handler_si
	add	r2, r2, #4
	ands r0, r1, #0x01 @ IRQ_VBLANK = BIT(0)
	bne	got_handler_mi
	add	r2, r2, #4
	ands r0, r1, #0x40 @ IRQ_TIMER3 = BIT(6)
	bne	got_handler_mi

@---------------------------------------------------------------------------------
IntrRet_nohand:
@---------------------------------------------------------------------------------
	str	r0, [r3, #0x0214]	@ IF Clear
	
	ldmfd   sp!, {r0,lr}	@ {spsr, lr}
	msr	spsr, r0		@ restore spsr
	mov	pc,lr

@---------------------------------------------------------------------------------
got_handler_si: @ single interrupt
@---------------------------------------------------------------------------------

	ldr	r1, [r2]		@ user IRQ handler address
	
	str	r0, [r3, #0x0214]	@ IF Clear

	adr	lr, IntrRet_si
	bx	r1

@---------------------------------------------------------------------------------
IntrRet_si:
@---------------------------------------------------------------------------------
	ldmfd   sp!, {r0,lr}	@ {spsr, lr}
	msr	spsr, r0		@ restore spsr
	mov	pc,lr

@---------------------------------------------------------------------------------
got_handler_mi: @ multiple interrupt
@---------------------------------------------------------------------------------

	ldr	r1, [r2]		@ user IRQ handler address
	
	mov	r2, #0
	str	r2, [r3, #0x208]	@ disable IME
	
	mrs	r2, cpsr
	bic	r2, r2, #0xdf		@ \__
	orr	r2, r2, #0x1f		@ /  --> Enable IRQ & FIQ. Set CPU mode to System.
	msr	cpsr,r2

	ldr	r2, [r3,#0x210]		@ REG_IE
	stmfd	sp!, {r0,r2, r3,lr}	@ irq mask, IE, REG_IE, lr
	bic	r2, r2, r0		@ disable interrupt about to be serviced
	str	r2, [r3,#0x210]

	mov	r2, #1
	str	r2, [r3, #0x208]	@ enable IME

	adr	lr, IntrRet_mi
	bx	r1

@---------------------------------------------------------------------------------
IntrRet_mi:
@---------------------------------------------------------------------------------
	mov	r3, #0x4000000		@ REG_BASE
	mov	r2, #0
	str	r2, [r3, #0x208]	@ disable IME

	ldmfd	sp!, {r0,r2, r3,lr}	@ irq mask, IE, REG_IE, lr
	str	r0, [r3, #0x0214]	@ IF Clear
	str	r2, [r3,#0x210]		@ restore REG_IE

	mrs	r3, cpsr
	bic	r3, r3, #0xdf		@ \__
	orr	r3, r3, #0x92		@ /  --> Disable IRQ. Enable FIQ. Set CPU mode to IRQ.
	msr	cpsr, r3

	mov	r3, #0x4000000		@ REG_BASE
	mov	r2, #1
	str	r2, [r3, #0x208]	@ enable REG_IME
	
	ldmfd   sp!, {r0,lr}	@ {spsr, lr}
	msr	spsr, r0		@ restore spsr
	mov	pc,lr

	.pool
	.end
