
#include <stdio.h>
#include <stdlib.h>

#include <NDS.h>

#include "emulator.h"
#include "_const.h"
#include "_console.h"
#include "filesys.h"
#include "memtool.h"

#include "mediatype.h"

#include "extmem.h"

enum EextmemType {EEMT_Unknown,EEMT_M3,EEMT_SC};

typedef struct {
  u32 StartAdr,EndAdr,CurAdr;
  u32 *pSlot;
  u32 SlotCount;
} TBody_M3;

typedef struct {
  u32 StartAdr,EndAdr,CurAdr;
  u32 *pSlot;
  u32 SlotCount;
} TBody_SC;

typedef struct {
  EextmemType Type;
  void *pBody;
} TInfo;

static TInfo *pInfo;

#define PacketSize_M3 (2)
#define PacketSize_SC (2)

// --- M3 stuff

static void SetM3_EnableCard(void) 
{
	// run unlock sequence
	volatile unsigned short tmp ;
	tmp = *(volatile unsigned short *)0x08000000 ;
	tmp = *(volatile unsigned short *)0x08E00002 ;
	tmp = *(volatile unsigned short *)0x0800000E ;
	tmp = *(volatile unsigned short *)0x08801FFC ;
	tmp = *(volatile unsigned short *)0x0800104A ;
	tmp = *(volatile unsigned short *)0x08800612 ;
	tmp = *(volatile unsigned short *)0x08000000 ;
	tmp = *(volatile unsigned short *)0x08801B66 ;
	tmp = *(volatile unsigned short *)0x08800006 ;
	tmp = *(volatile unsigned short *)0x08000000 ;
}

static void SetM3_EnablePSRAM(void) 
{
	// run unlock sequence
	volatile unsigned short tmp ;
	tmp = *(volatile unsigned short *)0x08000000 ;
	tmp = *(volatile unsigned short *)0x08E00002 ;
	tmp = *(volatile unsigned short *)0x0800000E ;
	tmp = *(volatile unsigned short *)0x08801FFC ;
	tmp = *(volatile unsigned short *)0x0800104A ;
	tmp = *(volatile unsigned short *)0x08800612 ;
	tmp = *(volatile unsigned short *)0x08000000 ;
	tmp = *(volatile unsigned short *)0x08801B66 ;
	tmp = *(volatile unsigned short *)0x08800004 ; // 0=bios?, 4 or c=8MByte?
	tmp = *(volatile unsigned short *)0x08000000 ;

        *(volatile u16*)0x09FFEFFE=0xAA55; // PepsiMan vote: The RAM on the M3 is made writable by writing 0xaa55 to 0x09ffeffe.

    // a=4; err=(n=0-7),(0x088nxxxx,0x089nxxxx,0x08anxxxx,0x08bnxxxx,0x08cnxxxx,0x08dnxxxx,0x08enxxxx,0x08fnxxxx)
}

// --- SC stuff

#define SSC_Disabled (0)
#define SSC_SDRAM (1)
#define SSC_CF (2)

#define SC_REG_UNLOCK	*(vu16*)(0x09FFFFFE)
static void SetSC_UNLOCK(int SSC)
{
  switch(SSC){
    case SSC_Disabled:
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0x0001;
      SC_REG_UNLOCK = 0x0001;
      break;
    case SSC_SDRAM:
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0x0005;
      SC_REG_UNLOCK = 0x0005;
      break;
    case SSC_CF:
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0x0003;
      SC_REG_UNLOCK = 0x0003;
  }
}
#undef SC_REG_UNLOCK

// ---

void extmem_Init(void)
{
  if(pInfo!=NULL) extmem_Free();
  
  pInfo=(TInfo*)safemalloc(sizeof(TInfo));
  MemSet32DMA3(0,pInfo,sizeof(TInfo));
  
  pInfo->Type=EEMT_Unknown;
  
  switch(DIMediaType){
    case DIMT_NONE: break;
    case DIMT_M3CF: pInfo->Type=EEMT_M3; break;
    case DIMT_M3SD: pInfo->Type=EEMT_M3; break;
    case DIMT_MPCF: break;
    case DIMT_SCCF: pInfo->Type=EEMT_SC; break;
    case DIMT_SCSD: pInfo->Type=EEMT_SC; break;
    case DIMT_NMMC: break;
    case DIMT_FCSR: break;
    case DIMT_SCS2: break;
    case DIMT_MPMD: break;
    case DIMT_EZSD: break;
  }
  
  switch(pInfo->Type){
    case EEMT_Unknown: {
      _consolePrintf("extmem:unknown cartridge type.\n");
    } break;
    case EEMT_M3: {
      _consolePrintf("extmem:set cartridge type to M3.\n");
      pInfo->pBody=(TBody_M3*)safemalloc(sizeof(TBody_M3));
      MemSet32DMA3(0,pInfo->pBody,sizeof(TBody_M3));
      
      TBody_M3 *pBody=(TBody_M3*)pInfo->pBody;
      
      pBody->StartAdr=0x08000000;
      pBody->EndAdr=pBody->StartAdr+(32*1024*1024)-8;
      pBody->CurAdr=pBody->StartAdr;
      pBody->pSlot=NULL;
      pBody->SlotCount=0;
    } break;
    case EEMT_SC: {
      _consolePrintf("extmem:set cartridge type to SC.\n");
      pInfo->pBody=(TBody_SC*)safemalloc(sizeof(TBody_SC));
      MemSet32DMA3(0,pInfo->pBody,sizeof(TBody_SC));
      
      TBody_SC *pBody=(TBody_SC*)pInfo->pBody;
      
      pBody->StartAdr=0x08000000;
      pBody->EndAdr=pBody->StartAdr+(32*1024*1024);
      pBody->CurAdr=pBody->StartAdr;
      pBody->pSlot=NULL;
      pBody->SlotCount=0;
    } break;
  }
}

void extmem_Free(void)
{
  if(pInfo==NULL) return;
  
  switch(pInfo->Type){
    case EEMT_Unknown: {
    } break;
    case EEMT_M3: {
      TBody_M3 *pBody=(TBody_M3*)pInfo->pBody;
      
      if(pBody->pSlot!=NULL){
        safefree(pBody->pSlot); pBody->pSlot=NULL;
      }
      
      safefree(pInfo->pBody); pInfo->pBody=NULL;
    } break;
    case EEMT_SC: {
      TBody_SC *pBody=(TBody_SC*)pInfo->pBody;
      
      if(pBody->pSlot!=NULL){
        safefree(pBody->pSlot); pBody->pSlot=NULL;
      }
      
      safefree(pInfo->pBody); pInfo->pBody=NULL;
    } break;
  }
  
  safefree(pInfo); pInfo=NULL;
}

void extmem_SetCount(u32 Count)
{
  if(pInfo==NULL) return;
  if(Count==0) return;
  
  switch(pInfo->Type){
    case EEMT_Unknown: {
    } break;
    case EEMT_M3: {
      TBody_M3 *pBody=(TBody_M3*)pInfo->pBody;
      
      if(pBody->pSlot!=NULL){
        _consolePrintf("extmem_SetSlotCount: Already allocated pSlot.\n");
        ShowLogHalt();
      }
      
      pBody->pSlot=(u32*)safemalloc(sizeof(u32)*Count);
      MemSet32DMA3(0,pBody->pSlot,sizeof(u32)*Count);
      pBody->SlotCount=Count;
    } break;
    case EEMT_SC: {
      TBody_SC *pBody=(TBody_SC*)pInfo->pBody;
      
      if(pBody->pSlot!=NULL){
        _consolePrintf("extmem_SetSlotCount: Already allocated pSlot.\n");
        ShowLogHalt();
      }
      
      pBody->pSlot=(u32*)safemalloc(sizeof(u32)*Count);
      MemSet32DMA3(0,pBody->pSlot,sizeof(u32)*Count);
      pBody->SlotCount=Count;
    } break;
  }
}

bool extmem_Exists(u32 SlotIndex)
{
  if(pInfo==NULL) return(false);
  
  switch(pInfo->Type){
    case EEMT_Unknown: {
      return(false);
    } break;
    case EEMT_M3: {
      TBody_M3 *pBody=(TBody_M3*)pInfo->pBody;
      
      if(pBody->SlotCount<=SlotIndex){
        _consolePrintf("extmem_Exists(u32 SlotIndex=%d); SlotCount==%d limit error.\n",SlotIndex,pBody->SlotCount);
        return(false);
      }
      
      if(pBody->pSlot[SlotIndex]==(u32)NULL) return(false);
    } break;
    case EEMT_SC: {
      TBody_SC *pBody=(TBody_SC*)pInfo->pBody;
      
      if(pBody->SlotCount<=SlotIndex){
        _consolePrintf("extmem_Exists(u32 SlotIndex=%d); SlotCount==%d limit error.\n",SlotIndex,pBody->SlotCount);
        return(false);
      }
      
      if(pBody->pSlot[SlotIndex]==(u32)NULL) return(false);
    } break;
  }
  
  return(true);
}

bool extmem_Alloc(u32 SlotIndex,u32 Size)
{
  if(pInfo==NULL) return(false);
  
  switch(pInfo->Type){
    case EEMT_Unknown: {
      return(false);
    } break;
    case EEMT_M3: {
      TBody_M3 *pBody=(TBody_M3*)pInfo->pBody;
      
      if(pBody->SlotCount<=SlotIndex){
        _consolePrintf("extmem_Alloc(u32 SlotIndex=%d,u32 Size=%d); SlotCount==%d limit error.\n",SlotIndex,Size,pBody->SlotCount);
        return(false);
      }
      
      Size=(Size+(PacketSize_M3-1)) & (~(PacketSize_M3-1));
      
      u32 TermAdr=pBody->CurAdr+Size;
      
      if(pBody->EndAdr<=TermAdr){
//        _consolePrintf("extmem_Alloc(u32 SlotIndex=%d,u32 Size=%d); MemAlloc limit error.\n",SlotIndex,Size);
        return(false);
      }
      
      // a=4; err=(n=0-7),(0x088nxxxx,0x089nxxxx,0x08anxxxx,0x08bnxxxx,0x08cnxxxx,0x08dnxxxx,0x08enxxxx,0x08fnxxxx)
      {
        u32 test=TermAdr>>16;
        if((0x0880<=test)&&(test<0x08f8)){
          if((test&0x000f)<0x0008){
            pBody->CurAdr=(TermAdr&0xfff00000)|0x00080000;
            TermAdr=pBody->CurAdr+Size;
          }
        }
      }
      
      pBody->pSlot[SlotIndex]=pBody->CurAdr;
      pBody->CurAdr=TermAdr;
    } break;
    case EEMT_SC: {
      TBody_SC *pBody=(TBody_SC*)pInfo->pBody;
      
      if(pBody->SlotCount<=SlotIndex){
        _consolePrintf("extmem_Alloc(u32 SlotIndex=%d,u32 Size=%d); SlotCount==%d limit error.\n",SlotIndex,Size,pBody->SlotCount);
        return(false);
      }
      
      Size=(Size+(PacketSize_SC-1)) & (~(PacketSize_SC-1));
      
      u32 TermAdr=pBody->CurAdr+Size;
      
      if(pBody->EndAdr<=TermAdr){
//        _consolePrintf("extmem_Alloc(u32 SlotIndex=%d,u32 Size=%d); MemAlloc limit error.\n",SlotIndex,Size);
        return(false);
      }
      
      pBody->pSlot[SlotIndex]=pBody->CurAdr;
      pBody->CurAdr=TermAdr;
    } break;
  }
  
  return(true);
}

bool extmem_Write(u32 SlotIndex,void *pData,u32 DataSize)
{
  if(pInfo==NULL) return(false);
  
  switch(pInfo->Type){
    case EEMT_Unknown: {
      return(false);
    } break;
    case EEMT_M3: {
      TBody_M3 *pBody=(TBody_M3*)pInfo->pBody;
      
      if(pBody->SlotCount<=SlotIndex){
        _consolePrintf("extmem_Write(u32 SlotIndex=%d,...); SlotCount==%d limit error.\n",SlotIndex,pBody->SlotCount);
        return(false);
      }
      
      u16 *pSrcData=(u16*)pData;
      u32 SrcSize=DataSize;
      u16 *pDstData=(u16*)pBody->pSlot[SlotIndex];
      u32 DstSize=(SrcSize+(PacketSize_M3-1)) & (~(PacketSize_M3-1));
      
      if(pDstData==NULL){
        _consolePrintf("extmem_Write(u32 SlotIndex=%d,...); pSlot[SlotIndex]==NULL not allocated error.\n",SlotIndex);
        return(false);
      }
      
      SetM3_EnablePSRAM();
/*
      for(u32 idx=0;idx<DstSize/2;idx++){
        pDstData[idx]=pSrcData[idx];
      }
*/
      MemCopy16DMA3(pSrcData,pDstData,DstSize);
      SetM3_EnableCard();
    } break;
    case EEMT_SC: {
      TBody_SC *pBody=(TBody_SC*)pInfo->pBody;
      
      if(pBody->SlotCount<=SlotIndex){
        _consolePrintf("extmem_Write(u32 SlotIndex=%d,...); SlotCount==%d limit error.\n",SlotIndex,pBody->SlotCount);
        return(false);
      }
      
      u16 *pSrcData=(u16*)pData;
      u32 SrcSize=DataSize;
      u16 *pDstData=(u16*)pBody->pSlot[SlotIndex];
      u32 DstSize=(SrcSize+(PacketSize_SC-1)) & (~(PacketSize_SC-1));
      
      if(pDstData==NULL){
        _consolePrintf("extmem_Write(u32 SlotIndex=%d,...); pSlot[SlotIndex]==NULL not allocated error.\n",SlotIndex);
        return(false);
      }
      
      SetSC_UNLOCK(SSC_SDRAM);
/*
      for(u32 idx=0;idx<DstSize/2;idx++){
        pDstData[idx]=pSrcData[idx];
      }
*/
      MemCopy16DMA3(pSrcData,pDstData,DstSize);
      SetSC_UNLOCK(SSC_CF);
    } break;
  }
  
  return(true);
}

bool extmem_Read(u32 SlotIndex,void *pData,u32 DataSize)
{
  if(pInfo==NULL) return(false);
  
  switch(pInfo->Type){
    case EEMT_Unknown: {
      return(false);
    } break;
    case EEMT_M3: {
      TBody_M3 *pBody=(TBody_M3*)pInfo->pBody;
      
      if(pBody->SlotCount<=SlotIndex){
        _consolePrintf("extmem_Read(u32 SlotIndex=%d,...); SlotCount==%d limit error.\n",SlotIndex,pBody->SlotCount);
        return(false);
      }
      
      u16 *pSrcData=(u16*)pBody->pSlot[SlotIndex];
//      u32 SrcSize=(Size+(PacketSize_M3-1)) & (~(PacketSize_M3-1));
      u16 *pDstData=(u16*)pData;
      u32 DstSize=DataSize;
      
      if(pSrcData==NULL){
        _consolePrintf("extmem_Read(u32 SlotIndex=%d,...); pSlot[SlotIndex]==NULL not allocated error.\n",SlotIndex);
        return(false);
      }
      
      SetM3_EnablePSRAM();
/*
      for(u32 idx=0;idx<DstSize/2;idx++){
        pDstData[idx]=pSrcData[idx];
      }
*/
      MemCopy16DMA3(pSrcData,pDstData,DstSize);
      SetM3_EnableCard();
    } break;
    case EEMT_SC: {
      TBody_SC *pBody=(TBody_SC*)pInfo->pBody;
      
      if(pBody->SlotCount<=SlotIndex){
        _consolePrintf("extmem_Read(u32 SlotIndex=%d,...); SlotCount==%d limit error.\n",SlotIndex,pBody->SlotCount);
        return(false);
      }
      
      u16 *pSrcData=(u16*)pBody->pSlot[SlotIndex];
//      u32 SrcSize=(Size+(PacketSize_SC-1)) & (~(PacketSize_SC-1));
      u16 *pDstData=(u16*)pData;
      u32 DstSize=DataSize;
      
      if(pSrcData==NULL){
        _consolePrintf("extmem_Read(u32 SlotIndex=%d,...); pSlot[SlotIndex]==NULL not allocated error.\n",SlotIndex);
        return(false);
      }
      
      SetSC_UNLOCK(SSC_SDRAM);
/*
      for(u32 idx=0;idx<DstSize/2;idx++){
        pDstData[idx]=pSrcData[idx];
      }
*/
      MemCopy16DMA3(pSrcData,pDstData,DstSize);
      SetSC_UNLOCK(SSC_CF);
    } break;
  }
  
  return(true);
}

